#!/usr/bin/python3
# -*- coding: utf-8 -*-
import subprocess
import json
import time

class ITDT:
    version = "0.3.0.20230228"
    def __init__(self, itdt_executable):
        self.itdt_executable = itdt_executable
        self._device_list = []
    

    def scan_devices(self, verbose=True):
        cmd = [self.itdt_executable, 'scanj']
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        data, error = proc.communicate()
        #print('Data: '  + data.decode('ascii'))
        #print('Error: '  + error.decode('ascii'))
        #print('code: ' + str(proc.returncode))
        if proc.returncode == 0:
            self._device_list = json.loads(data)

        if verbose and proc.returncode == 0:
            for device in self._device_list:
                #print(device['modelName'] + " --> " + device['serialNumber'])
                print(device)
        
        return self._device_list        

    def run_commands(self, device, commands, return_json=False):
        cmds = []
        cmds.append(self.itdt_executable)
        cmds.append('-f')
        cmds.append(device)
        cmds += commands.split()
        #cmds = [self.itdt_executable, '-f', device, 'inqj', '0x83']
        proc = None
        data = None
        try:
            proc = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            data, error = proc.communicate()
            if proc.returncode == 0:
                if data and return_json:
                    data = json.loads(data)
            if error:
                data = error.decode('ascii')      
            
        except (RuntimeError, TypeError, NameError):
            return -1, None

        return proc.returncode, data
    
    
    def move_cartridge(self, device, hdc, source, destination, verbose=False):
        #print('Move')
        # ITDT's new move command supporting hdc parameter
        command=f'move -hdc {hdc} {source} {destination}'
        if verbose:
            print("Move command: ", command)

        start_time = time.time()        
        result, data = self.run_commands(device, command)
        diff_time = time.time() - start_time
        if result == 0:
            if verbose:
                print(data)
        else:   
            print(f"Error while move cartridge {source}")

        return result, diff_time



class ITDTROS(ITDT):

    def __init__(self, itdt_executable, device_file_changer, device_file_drive=None):
        ITDT.__init__(self, itdt_executable)
        self.cartridges = []
        self.cartridge_count=0
        self.slot = {}
        self.device_changer = device_file_changer
        self.device_drive   = device_file_drive
        self.tier=9
    
    def load_slot_info(self, slot, verbose=False):
        #print('Load slot info ' + slot)
        result, data = self.run_commands(self.device_changer, 'ros GET /v1/slots/'+ slot, True)
        if result == 0:
            self.cartridges = data[0]['contents']
            self.slot = data[0] 
            self.cartridge_count=len(self.cartridges)-self.cartridges.count(None)           
            self.tier = len(self.cartridges)
            if verbose:
                print(f"Cartridge count for slot: {slot} is {self.cartridge_count}")
        else:   
            print(f"Error while reading slot info {slot}")            

        return result, data


    def load_data_cartridge_info(self, volser, verbose=False):
        #print('Load volser info ' + volser)
        result, data = self.run_commands(self.device_changer, 'ros GET /v1/dataCartridges/'+ volser, True)
        if verbose:
            print(data)            
        
        return result, data

    
    def get_drive_info(self, serial, verbose=False):
        if not isinstance(serial, str):
            return -1, None
        result, data = self.run_commands(self.device_changer, 'ros GET /v1/drives/'+ serial, True)
        if verbose:
            print(data)            
        
        return result, data

    
    def check_slot(self, logical_library, slot):
        result, data = self.load_slot_info(slot)
        if result != 0:
            #print ("Load slot info failed")
            return result, None
        
        if self.cartridge_count >= 0:
            count = 0            
            for volser in self.cartridges:
                if volser is None:
                    continue
                result, data = self.load_data_cartridge_info(volser)
                if result == 0 and len(data) == 1:
                    #print(f"Volser: {volser}, Elementaddress: {data[0]['elementAddress']}, internalAddress: {data[0]['internalAddress']}")
                    ll = data[0]['logicalLibrary']
                    if ll != logical_library:
                        print (f"LogicalLibray {logical_library} does not match for {data[0]}")
                        count+=1
            if count > 0:
                print ("!!! Either select an other slot or move the cartridges.")                        
            return count


    def find_data_cartridges(self, logical_library):
        result, data = self.run_commands(self.device_changer, 'ros GET /v1/dataCartridges', True)
        ret = []
        if result == 0:            
            #print (data)            
            for cartridge in data:
                #print (cartridge)
                if cartridge['logicalLibrary'] == logical_library and cartridge['state'] == 'normal' \
                and (cartridge['location'].endswith('T1') or cartridge['location'].endswith('T2')):
                    if cartridge['volser'] not in self.cartridges:
                        ret.append(cartridge['volser'])
            return ret
        else:
            return None        

    
    def move_to_slot(self, slot, volume, verbose=False):
        # Check first, if the tier is 1 or 2, otherwise the moveToSlot will fail
        result, data = self.load_data_cartridge_info(volume, False)
        if result == 0:
            location = data[0]['location']
            if location.endswith('T1') or  location.endswith('T2'):
                parameter = '{"type":"moveToSlot", "cartridge":"' + volume + '", "destinationLocation":"' + slot + '"}'
                result, data = self.run_commands(self.device_changer, 'ros POST /v1/workItems ' + parameter, True)
                ret = []
                if result == 0 and verbose:
                    print(f"Cartridge: {volume} moved to slot: {slot}")
                else:
                    print(f"Move failed: {data}")

    def prepare_slot_for_test(self, logical_library, slot, verbose=False, headless=False):
        if self.check_slot(logical_library, slot) == 0:     # Check for cartridges from other LL's 
            self.load_slot_info(slot)                       # Get the numbers for a dedicated slot
            print(f"Slot: {slot} has {self.cartridge_count} cartridge(s) loaded")
            current_cartridges = self.cartridge_count
            if self.cartridge_count < self.tier:
                answer='y'
                #print(f'Slot: {slot}, cartridges: {self.cartridges}')
                if not headless:            
                    answer = input(f'Should be other cartridges loaded to slot {slot} to fill up(y/n)? ')
                if answer.casefold()=="y":
                    print ('Loading cartridges to slot: ', slot)
                    carts = self.find_data_cartridges(logical_library)
                    missing = self.tier - self.cartridge_count
                    if len(carts) > missing:
                        print (carts)
                        #Moving cartridges into slot - fill up all tiers
                        for x in range(0, missing):
                            self.move_to_slot(slot+'T1', carts[x], verbose)
                            current_cartridges += 1
                            time.sleep(1)
                    else:
                        print("No cartridge available (in TIER 0/1) for logical library {logical_library} to get loaded into {slot}")
                                
                # Synchronize the move_to_slot calls - wait till move is completed
                loop_counter = 10
                while (current_cartridges > self.cartridge_count) and (loop_counter > 0):
                    time.sleep(5)
                    self.load_slot_info(slot)   #update numbers in class members    
                    loop_counter -= 1
        else:
            print(f"Slot {slot} is fully loaded.")

    def mount_and_unmount_cartridge(self, drive, cart_number=None, hdc=None):
        while cart_number is None:
            cart_number = int(input(f'Select cartridge number to load 1 to {self.cartridge_count}: '))
            if (cart_number < 1 and cart_number > self.cartridge_count):
                cart_number=None
        
        while hdc not in range(3):
            hdc = int(input('Select hdc value for move medium (0|1|2|3): '))        
        if hdc < 0:
            hdc=0

        volume = self.cartridges[cart_number-1]
        
        result, cartridge = self.load_data_cartridge_info(volume)
        if result == 0:            
            source_address = cartridge[0]['elementAddress']
            drive_element_address = drive['elementAddress']
            drive_sn = drive['sn']
            print (f'  Moving cartridge {volume}[{source_address}] to drive {drive_sn}[{drive_element_address}]...', end='',  flush=True)
            result,mount = self.move_cartridge(self.device_changer, hdc, source_address, drive_element_address, False)
            print(f'{mount:.2f}')
            if result == 0:
                if self.device_drive:
                    print(f'  Unloading cartridge from drive {drive_sn}...', end='', flush=True)
                    self.run_commands(self.device_drive, 'load tur tur') #Catching sk/asc/ascq 06/28/00 ...MEDIUM MAY HAVE CHANGED 
                    result, data = self.run_commands(self.device_drive, 'unload')
                    if result == 0:
                        print('Done')
                    else:
                        print(f'Failed with {result}--> {data}')
                
                print (f'  Moving cartridge {volume}[{source_address}] from drive {drive_sn}[{drive_element_address}] back...', end='', flush=True)
                result,umount = self.move_cartridge(self.device_changer, 0, drive_element_address, source_address, False)
                print(f'{umount:.2f}')
                if result != 0:
                    print(f'Move back failed - {drive_element_address} to {source_address}')    
                    return -2, None, None
            else:
                print(f'Move failed - {source_address} to {drive_element_address}')
                return -1, None, None
            return 0, mount, umount

